//**********************
// getIR_nakata.c for Raspberry Pi
//
// (1)NEC format
// IR carrier=38KHz
// Time unit=0.56ms
// logical 0 = on 1T + off 1T
// logical 1 = on 1T + off 3T
// reader=on 16T + off 8T
// stop=on 1T
// frame=108ms
//
// (2)AEHA format
// IR carrier=33 - 40KHz
// Time unit=0.35 - 0.50ms
// logical 0 = on 1T + off 1T
// logical 1 = on 1T + off 3T
// reader=on 8T + off 4T
// trailer=on 1T + 8ms
//
// (3)SONY format
// IR carrier=40KHz
// Time unit=0.6ms
// logical 0 = off 1T + on 1T
// logical 1 = off 1T + on 2T
// reader=on 4T
// frame=45ms
//
// caution:
// no detecting repeat code, return bits=0;
//
// (C)Copyright 2015 All rights reserved by Y.Onodera
// http://einstlab.web.fc2.com
//
// modified part
// Copyright (c) 2019 by NAKATA H. All rights reserved
// http://www.nakata-jp.org/
//**********************

#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <bcm2835.h>


// with ApplePi board, you must comment out below 1 line definition
#define	NEED_PULLUP	1			// IR sensor pin needs pull up

#define	IR_LIMITS	1024		// bytes buffer = IR_LIMITS x8 bits
#define	MILLION(x)	((x) * 1000000)		// x * 1000 * 1000
#define	TIME_OUT	10			// wait 10 sec

#define	PIN_IRIN	RPI_V2_GPIO_P1_07	// irin  = GPIO4  = P1_7
#define	CYCLE_38kHz	26			// 26 milli sec in 1 cycle
#define	BITS_IN_BYTE	8			// 8 bits in 1 byte
#define	TRUE		(-1)			// boolean like value
#define	FALSE		0			// boolean like value
#define	WAIT_CYCLE(t)	bcm2835_delayMicroseconds((t) * CYCLE_38kHz)
#define	IS_PIN_HIGH	( bcm2835_gpio_lev(PIN_IRIN) == 1 )
#define	IS_PIN_LOW	( bcm2835_gpio_lev(PIN_IRIN) == 0 )

static enum type_e_mode {E_UNKNOWN, E_NEC, E_AEHA, E_SONY} e_mode;
static uint8_t		buf[IR_LIMITS];		// buffer of bit flags
static unsigned int	total_bits = 0;		// 32768 bits capable
static char		*error_message = NULL;


static char	*mode2name(enum type_e_mode mode)
{
	switch (mode) {
	case E_UNKNOWN:
		return "none";
	case E_NEC:
		return "NEC ";
	case E_AEHA:
		return "AEHA";
	case E_SONY:
		return "SONY";
	}
	// sequence must not reach here
}

static uint8_t	countIROn()
{
	unsigned int		i;
	const unsigned int	limit_count = 256;
	const time_t		tm_start = time(NULL);

	while (IS_PIN_HIGH) {
		const time_t	tm_now = time(NULL);
		if (tm_now - tm_start > TIME_OUT) {
			error_message = "IR Off duration is too long.\n";
			return 0;	// too long OFF status
		}
	}

	for (i = 0; IS_PIN_LOW; i++ ){
		WAIT_CYCLE(2);
		if ( i >= limit_count ) {
			error_message = "IR On duration is too long.\n";
			return 0;	// too long ON status
		}
	}
	// NEC:i=19*8=152, i*2*26.5us=8056us
	// AEHA:i=19*4=76, i*2*26.5us=4028us
	// 1T:i=19*1=19

	return i;
}


static uint8_t	countIROff()
{
	unsigned int		i;
	const unsigned int	limit_count = 256;
	const time_t		tm_start = time(NULL);

	while (IS_PIN_LOW) {
		const time_t	tm_now = time(NULL);
		if (tm_now - tm_start > TIME_OUT) {
			error_message = "IR On duration is too long.\n";
			return 0;	// too long ON status
		}
	}

	for ( i = 0; IS_PIN_HIGH; i++ ) {
		WAIT_CYCLE(1);
		if (i >= limit_count) {
			error_message = "IR Off duration is too long.\n";
			return 0;	// too long OFF status
		}
	}
	// NEC:i=19*8=152, i*26.5us=4028us
	// AEHA:i=19*4=76, i*26.5us=2014us
	// 1T:i=19*1=19
	// 3T:i=19*3=57

	return i;
}


static int	getIR2()
{
	uint16_t	i;	// must be 16bit in order to loop end check
	uint16_t	j;	// capable 32768 bits = 4096 bytes

	total_bits = 0;
	for (j = 0; j < IR_LIMITS; j++) {		// buffer bytes LIMITS
		buf[j] = 0;				// clear flags
		for (i = 1; i <= 0x80; i <<= 1) {	// bit mask operation
			uint8_t	on_duration = countIROn();
			uint8_t	off_duration = countIROff();
			if (e_mode == E_SONY) {
				// Threschold = 35, 23 = 1T, 46 = 2T; for SONY
				if (on_duration > 15) {
					buf[j] |= i;	// set flag
				}
				++total_bits;
				if (off_duration == 0) {
					return TRUE;	// end of SONY pattern
				}
			} else {
				if (off_duration == 0) {
					return TRUE; // end of AEHA,NEC pattern
				}
				// Threschold = 38, 19 = 1T, 57 = 3T; for NEC
				// Threschold = 30, 15 = 1T, 45 = 3T; for AEHA
				if (off_duration > 30) {
					buf[j] |= i;	// set flag
				}
				++total_bits;
			}
		}
	}
	error_message = "input buffer over flow. pattern is too long\n";
	return FALSE;		// buffer over flow
}


static int	getIR()
{
	unsigned int	i;

	// to detect falling edge
	for (i = 0; IS_PIN_HIGH; i++) {
		if (i >= MILLION(TIME_OUT)) {
			error_message = "Waiting IR off timed out\n";
			return FALSE;		// delay is too long, give up
		}
		bcm2835_delayMicroseconds(1);	// short wait
	}

	e_mode = E_UNKNOWN;
	i = countIROn();
	if (i == 0) {
		error_message = "Waiting IR pattern start timed out\n";
		return FALSE;		// Start
	}
	if (i <= 40) {
		error_message = "IR On duration is too short on starting.\n";
		return FALSE;	// too short intermission
	} else if (i < 51) {
		e_mode = E_SONY;	// SONY, 46
	} else if (i <= 100) {
		e_mode = E_AEHA;	// AEHA, 54-77
	} else {
		e_mode = E_NEC;	// NEC, 173
	}
	if (countIROff() == 0) {		// skip signal off
		error_message = "Waiting IR 2nd code timed out\n";
		return FALSE;
	}
	return getIR2();
}

static int	init_pin()
{
	if (!bcm2835_init()) {
		fprintf(stderr, "SoC initialization failed.\n");
		return FALSE;
	}

	bcm2835_gpio_fsel(PIN_IRIN,  BCM2835_GPIO_FSEL_INPT);
#ifdef	NEED_PULLUP
	bcm2835_gpio_set_pud(PIN_IRIN,  BCM2835_GPIO_PUD_UP);
#endif

	return TRUE;
}

static void	usage(const char *command_name)
{
	fprintf(stderr, "usage:\t%s [loop_times]\n", command_name);
}

int	main(int argc, char *argv[])
{
	int	i;
	int	loop_count = 1;

	if (geteuid() != 0) {
		fprintf(stderr, "please execute as superuser.\n");
		return 1;
	}

	if (argc == 2) {
		if (sscanf(argv[1], "%d", &loop_count) != 1) {
			fprintf(stderr, "argument %s must be a number!\n", argv[1]);
			return 1;
		}
	}

	if (!init_pin())
		return 1;

	while (loop_count-- > 0) {
		if (!getIR()) {
			fprintf(stderr, "%s\ndetection failed.\n",
				error_message);
			return 1;
		}

		// print out result
		printf("%s %d", mode2name(e_mode), total_bits);
		for (i = 0; i < total_bits; i += BITS_IN_BYTE) {
			printf(" %d", buf[i / BITS_IN_BYTE]);	// data
		}
		printf("\n");
	}

	return 0;
}
