//**********************
// setIR_nakata.c for Raspberry Pi
//
// (1)NEC format
// IR carrier=38KHz
// Time unit=0.56ms
// logical 0 = on 1T + off 1T
// logical 1 = on 1T + off 3T
// reader=on 16T + off 8T
// stop=on 1T
// frame=108ms
//
// (2)AEHA format
// IR carrier=33 - 40KHz
// Time unit=0.35 - 0.50ms
// logical 0 = on 1T + off 1T
// logical 1 = on 1T + off 3T
// reader=on 8T + off 4T
// trailer=on 1T + 8ms
//
// (3)SONY format
// IR carrier=40KHz
// Time unit=0.6ms
// logical 0 = off 1T + on 1T
// logical 1 = off 1T + on 2T
// reader=on 4T
// frame=45ms
//
// caution:
// no detecting repeat code, return bits=0;
//
// (C)Copyright 2015 All rights reserved by Y.Onodera
// http://einstlab.web.fc2.com
//
// modified part
// Copyright (c) 2019 by NAKATA H. All rights reserved
// http://www.nakata-jp.org/
//**********************

#include <stdio.h>
#include <string.h>
#include <bcm2835.h>


#define	IR_LIMITS	64  // bytes buffer = IR_LIMITS x8 bits
#define	AEHA	19
#define	NEC	22
#define	SONY	23

typedef enum { E_NONE, E_NEC, E_AEHA, E_SONY } E_PATTERN;
static E_PATTERN	e_pattern = E_NONE;

static void outON(char n, char t);
static void outOFF(char n, char t);
static void setIR();

static unsigned char buf[IR_LIMITS];  // bytes buffer
static unsigned short bits;    // 32768 bits capable



// irin  = GPIO4  = J8_7
// irout = GPIO13 = J8_33
#define IROUT RPI_BPLUS_GPIO_J8_33


// out ON with 38KHz
void outON(char n, char t)
{

	unsigned char i,j;

	for(j=0;j<t;j++){
		for(i=0;i<n;i++){
			// 38KHz, 1/3duty
			bcm2835_gpio_write(IROUT, HIGH); // LED ON=8.6ms
			bcm2835_delayMicroseconds(9);
			bcm2835_gpio_write(IROUT, LOW);	// LED OFF=17.4ms
			bcm2835_delayMicroseconds(17);
		}
	}

}

// out OFF without 38KHz
void outOFF(char n, char t)
{

	unsigned char i,j;

	for(j=0;j<t;j++){
		for(i=0;i<n;i++){
			// 38KHz, 1/3duty
			bcm2835_gpio_write(IROUT, LOW); // LED OFF
			bcm2835_delayMicroseconds(9);
			bcm2835_gpio_write(IROUT, LOW); // LED OFF
			bcm2835_delayMicroseconds(17);
		}
	}

}


void	setIR()
{

	unsigned char j,t;
	unsigned short i,n;

	if (bits == 0) {
		return;      // no data
	}

	// reader
	switch(e_pattern){
	case E_NEC:
		if (bits != 32) return;
		outON(NEC,16);  // ON(16T)
		outOFF(NEC,8);  // OFF(8T)
		t = 16 + 8;
		break;
	case E_AEHA:
		if (bits > IR_LIMITS * 8) {
			return;
		}
		outON(AEHA,8);  // ON(8T)
		outOFF(AEHA,4); // OFF(4T)
		t = 8 + 4;
		break;
	case E_SONY:
		if( !(bits==12 || bits==15 || bits==20)) {
			return;
		}
		outON(SONY,4);  // ON(4T)
		t = 4;
		break;
	}

	// data
	switch(e_pattern){
	case E_NEC:
		for (i = 0; i < 4; i++) {
			int bit_pattern = buf[i];
			for (j = 0; j < 8; j++) {
				outON(NEC,1);   // ON(1T)
				if (bit_pattern & 1) {
					outOFF(NEC,3);  // OFF(3T)
					t+=4;
				} else {
					outOFF(NEC,1);  // OFF(1T)
					t+=2;
				}
				bit_pattern >>= 1;
			}
		}
		break;
	case E_AEHA:
		i = 0;
		n = 0;
		do {
			int bit_pattern = buf[i];
			for (j = 0; j < 8; j++) {
				outON(AEHA,1);  // ON(1T)
				if (bit_pattern & 1) {
					outOFF(AEHA,3); // OFF(3T)
					t += 4;
				} else {
					outOFF(AEHA,1); // OFF(1T)
					t += 2;
				}
				if (++n == bits)
					break;
				bit_pattern >>= 1;
			}
			++i;
		} while (n < bits && i < IR_LIMITS);
		break;
	case E_SONY:
		i = 0;
		n = 0;
		do {
			int bit_pattern = buf[i];
			for(j = 0; j < 8; j++) {
				outOFF(SONY,1); // OFF(1T)
				if (bit_pattern & 1) {
					outON(SONY,2);  // ON(2T)
					t += 3;
				} else {
					outON(SONY,1);  // ON(1T)
					t += 2;
				}
				if (++n == bits)
					break;
				bit_pattern >>= 1;
			}
			++i;
		} while(n < bits && i < IR_LIMITS);
		break;
	default:
		break;	// do nothing
	}
   
	// stop
	switch(e_pattern){
	case E_NEC:
		t = 192 - t;
		outON(NEC, 1);   // ON(1T)
		outOFF(NEC, t);  // frame=108ms=192T
		break;
	case E_AEHA:
		outON(AEHA, 1);  // ON(1T)
		outOFF(AEHA, 16); // 8ms=16T
		break;
	case E_SONY:
		t = 75 - t;
		outOFF(SONY, t); // frame=45ms=75T
		break;
	default:
		break;	// do nothing
	}
}

typedef struct {
	const E_PATTERN	pattern;
	const char	*name;
} s_pattern;
static const s_pattern	s_patterns[] = {
	{E_NEC, "NEC"}, {E_AEHA, "AEHA"}, {E_SONY, "SONY"}, {E_NONE, NULL}
};

int main(int argc, char *argv[])
{
	char i,n;

	// check parameter
	if ( argc < 2 ) {
		fprintf(stderr, "too few args\n");
		return 1;
	}

	for ( i = 0; s_patterns[i].name != NULL; i++) {
		if (strcmp(argv[1], s_patterns[i].name) == 0) {
			break;
		}
	}
	if (s_patterns[i].name == NULL) {
		fprintf(stderr, "pattern %s not defined.\n", argv[1]);
		return 1;
	}
	e_pattern = s_patterns[i].pattern;


 	bits = atoi(argv[2]);
	if ( bits < 1 || 32768 < bits ) {
		fprintf(stderr, "Illegal bits %d\n", bits);
		return 1;
	}

	// set data
	n = (bits+7)/8;    // bits to n bytes
	if ( argc != (n+3) ) return 1;
	for(i=0;i<n;i++){
	 	buf[i] = atoi(argv[3+i]);
	}


	if (!bcm2835_init()) {
		fprintf(stderr, "bcm2835 initialize failed.\n");
		return 1;
	}

	bcm2835_gpio_fsel(IROUT, BCM2835_GPIO_FSEL_OUTP);
	bcm2835_gpio_write(IROUT, LOW);

	// set parameter to send out
	// mode is 1:NEC, 2:AEHA or 3:SONY
	// bits is number to send out less than IR_LIMITS x8
	// buf[] are bytes buffer

	setIR();
	setIR();
	setIR();

	bcm2835_gpio_write(IROUT, LOW);

	return 0;

}
