#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>

#ifdef	WIRING_PI
#include <wiringPi.h>
#define	PIN_IRIN	7
#define PIN_IROUT	23
#define	IS_PIN_HIGH	( digitalRead(PIN_IRIN) == 1 )
#define	IS_PIN_LOW	( digitalRead(PIN_IRIN) == 0 )
#define	OUT_ON		digitalWrite(PIN_IROUT, 1)
#define	OUT_OFF		digitalWrite(PIN_IROUT, 0)
#define	USLEEP(time)	delayMicroseconds(time)
#else	// WIRING_PI
#include <bcm2835.h>
#define	PIN_IRIN	RPI_V2_GPIO_P1_07	// infra red in  = GPIO4  = P1_7
#define PIN_IROUT	RPI_BPLUS_GPIO_J8_33	// infra red out
#define	IS_PIN_HIGH	( bcm2835_gpio_lev(PIN_IRIN) == 1 )
#define	IS_PIN_LOW	( bcm2835_gpio_lev(PIN_IRIN) == 0 )
#define	OUT_ON		bcm2835_gpio_write(PIN_IROUT, HIGH)
#define	OUT_OFF		bcm2835_gpio_write(PIN_IROUT, LOW)
#define	USLEEP(time)	bcm2835_delayMicroseconds(time)
#endif	// WIRING_PI


#define	MILLION(x)	((x) * 1000000)		// x * 1000 * 1000
#define	TIME_OUT	10			// wait 10 sec
#define	THRESHOLD	10000			// 10 mili sec
#define	CYCLE_38kHz	26			// 26 milli sec in 1 cycle
#define	BUFF_LEN	1024

#define	TRUE		(-1)			// boolean like value
#define	FALSE		0			// boolean like value

enum E_CONTROL {E_ON, E_OFF};

static char		*error_message = NULL;


static int	init_pin()
{
#ifdef	WIRING_PI
	wiringPiSetup();

	pinMode(PIN_IROUT, OUTPUT);
	pinMode(PIN_IRIN, INPUT);
	pullUpDnControl(PIN_IRIN, PUD_UP);
#else	// WIRING_PI
	if (!bcm2835_init()) {
		fprintf(stderr, "SoC initialization failed.\n");
		return FALSE;
	}

        bcm2835_gpio_fsel(PIN_IROUT, BCM2835_GPIO_FSEL_OUTP);
	bcm2835_gpio_fsel(PIN_IRIN,  BCM2835_GPIO_FSEL_INPT);
	bcm2835_gpio_set_pud(PIN_IRIN,  BCM2835_GPIO_PUD_UP);
#endif	// WIRING_PI

	OUT_OFF;

	return TRUE;
}

static inline int	subtract_tv(struct timeval *p_before, 
					struct timeval *p_after)
{
	int	diff;

	diff = MILLION(p_after->tv_sec - p_before->tv_sec);
	diff += (p_after->tv_usec - p_before->tv_usec);

	return diff;
}

#ifdef	DEBUG
static inline void	print_val(char *str, int timing)
{
#ifdef	DEBUG
	const int	val = timing / CYCLE_38kHz;
	const int	rest = timing % CYCLE_38kHz;

	printf("%s %d.%d", str, val, rest);
#else	// DEBUG
	printf("%s %d", str, timing);
#endif	// DEBUG
}
#endif	// DEBUG

static int	receive()
{
	int	i;
	int	loop_count = 1;
	struct timeval	tv_up_edge;
	struct timeval	tv_down_edge;
	int	off_duration;

	off_duration = -1;	// set start mark
	gettimeofday(&tv_up_edge, NULL);
	while (IS_PIN_HIGH) {
		gettimeofday(&tv_up_edge, NULL);
	}
	while (1) {
		while (IS_PIN_LOW) {
			gettimeofday(&tv_down_edge, NULL);
		}
		if (off_duration < 0) {
			print_val("", subtract_tv(&tv_up_edge, &tv_down_edge));
		} else {
			printf_val( (off_duration <= THRESHOLD) ? ", " : "\n\n",
					-off_duration);
			printf_val(",", 
				subtract_tv(&tv_up_edge, &tv_down_edge));
		}
		fflush(stdout);
		while (IS_PIN_HIGH) {
			gettimeofday(&tv_up_edge, NULL);
			if (tv_up_edge.tv_sec - tv_down_edge.tv_sec > 
								TIME_OUT) {
				printf("\n");
				return 0;	// too long silence detected
			}
		}
		off_duration = subtract_tv(&tv_down_edge, &tv_up_edge);
	}

	exit(1);	// sequenct must not reach here
}

static int	output_pulses(E_CONTROL e_control, int rest_count)
{
	while (rest_count > 0) {
		if (E_CONTROL == E_ON) {
			OUT_ON;		// ON duty is 1/3
		} else {
			OUT_OFF;	// keep OFF
		}
		USLEEP(9);		// sleep 9 micro sec
		OUT_OFF;		// OFF duty is 2/3
		USLEEP(17);		// sleep 17 micro sec
		rest_count -= 26;	// 26 micro sec passed
	}

	retrun rest_count;
}

static int	send(int argc, char *argv[])
{
	int		i;
	int		timing[BUFF_LEN];
	int		count = 0;
	const char	*p_param = strdup(argv[1]);
	char		*p_start = p_param;
	char		*p_end;
	int		rest_count = 0;

	OUT_OFF;

	if (!p_param) {
		fprintf(stderr, "strdup() failed.\n");
		exit(1);
	}

	for (i = 0; i < BUFF_LEN; i++) {
		timing[i] = 0;
	}

	// parse parameter
	for (p_end = p_param; *p_end; p_end++) {
		if (*p_end != ',') {
			continue;
		}

		// parse integer before ','
		if (count >= BUFF_LEN) {
			fprintf(stderr, "buffer overflow.\n");
			exit(1);
		}
		*p_end = '\0';
		if (sscanf(p_start, "%d", timing + count) != 1) {
			fprintf(stderr, "illegal parameter %s\n", p_start);
			exit(1);
		}
		++count;

		p_start = p_end + 1;
	}
	if (p_start != p_end) {
		// read last integer
		if (count >= BUFF_LEN) {
			fprintf(stderr, "buffer overflow.\n");
			exit(1);
		}
		if (sscanf(p_start, "%d", timing + count) != 1) {
			fprintf(stderr, "illegal parameter %s\n", p_start);
			exit(1);
		}
		++count;
	}
	free(p_param);	// parse end.

	// send IR signals
	for (i = 0; i < count;) {
		rest_count += timing[i++];
		if (rest_count <= 0) {
			fprintf(stderr, "illegal parameter %s\n", argv[1]);
			exit(1);
		}
		// ON cycle
		rest_count = output_pulses(E_ON, rest_count);

		rest_count += timing[i++];
		if (rest_count == 0) {
			break;			// end of signals
		}
		if (rest_count < 0) {
			fprintf(stderr, "illegal parameter %s\n", argv[1]);
			exit(1);
		}
		// OFF cycle
		rest_count = output_pulses(E_OFF, rest_count);
	}

	OUT_OFF;
	return 0;
}

int	main(const int argc, const char *argv[])
{
	if (geteuid() != 0) {
		fprintf(stderr, "please execute as superuser.\n");
		return 1;
	}

	if (!init_pin()) {
		return 1;
	}

	if (argc == 1) {
		return receive();
	} else {
		return send(argc, argv);
	}
}
